/*-------------------------------------------------------------------------* 

Title: Code to implement the adaptive threshold outlier detection method   * 
*--------------------------------------------------------------------------* 
Author:  Paulo Macedo                                                      * 
Created: March 01, 2016                                                    * 

Description: Macro Mdist reads the input dataset, which has a record per 
provider Records include information on the provider identifier, the 
cluster number to which the provider was assigned and the variables 
required for computing distance measures from the sampled data points to 
the center of their distribution - those will be the right-hand  side  
variables to be  entered in the model statement of PROC ROBUSTREG 

Parameters:
   cluster= the cluster number identifying the subset of the input file
            that will be analyzed; the value "ALL" is an admissible entry
            that will analyze the whole dataset instead of any particular
            cluster;
   input=   the name of the input dataset;
   seed=    assumes any value from the set of integer numbers to assure 
            the possibility of replicating results; 
   p=       the number of variables in the analysis.
*-------------------------------------------------------------------------*/ 

OPTIONS MLOGIC SYMBOLGEN;

%macro ODSOff(); /* Called prior to BY-group processing */
ods graphics off;
ods exclude all;
ods noresults;
%mend;
 
%macro ODSOn(); /* Called after BY-group processing */
ods graphics on;
ods exclude none;
ods results;
%mend;

%macro Mdist(cluster, input, seed, p);

%if %UPCASE(&cluster) = ALL %then %do;
Data Cluster_&cluster;
Set &input;
run;
%end;
%else %do;
Data Cluster_&cluster;
Set &input;
if CLUSTER = &cluster;
run;
%end;

Data Cluster_&cluster;
set Cluster_&cluster; 
format Log_amount comma12.2;
call streaminit(&seed); 
z = rand('normal');
run;

/* query saves infomation about sample size to macro variable n */
Proc SQL Noprint;
select Count(*) into :n
from Cluster_&cluster;
quit;

%put &n;

%ODSoff;
proc robustreg data=Cluster_&cluster method=MM;
model z = Amount_per_serv  total_payment_hcpcs  
/leverage;
output out = hat_robust_m RD = RD_DIST MD = Mahal ;
run;
%ODSon;

Proc sort data = hat_robust_m; 
by descending mahal; run;

Data Mhl_distance_detection_&cluster;
set Hat_robust_m;
Format DISTRIBUTION_STATUS $15.;
DISTRIBUTION_STATUS = "WITHIN THE NORM";
if PROBCHI((Mahal)**2, &p) > 0.975 then DISTRIBUTION_STATUS = "OUTLIER";
run;

Proc sort data = hat_robust_m out = hat_robust_m_v2; 
by descending RD_DIST; run;

Data Robust_distance_detection_&cluster;
set Hat_robust_m_v2;
Format DISTRIBUTION_STATUS $15.;
DISTRIBUTION_STATUS = "WITHIN THE NORM";
if PROBCHI((RD_DIST)**2, &p) > 0.975 then DISTRIBUTION_STATUS = "EXTREME VALUE";
run;

proc sort data = Robust_distance_detection_&cluster; by descending RD_DIST; run;

/*********************************************************************************
THE ADAPTIVE OUTLIER DETECTION METHOD WAS PROPOSED BY DANIEL GERVINI (2002) AND
REINMANN, FILZMOSER AND GARRET (2005); IT USES A CRITICAL THRESHOLD TO SEPARATE 
EXTREME VALUES OF THE MAIN DISTRIBUTION ASSOCIATED WITH THE DATA GENERATING PROCESS 
FROM OUTLIER VALUES DRAWN (POSSIBLY) FROM OTHER(S) (CONTAMINATING) DISTRIBUTION(S) 
**********************************************************************************/

/**********************************************************************
THE IDENTIFICATION OF THE CRITICAL THRESHOLD REQUIRES THE COMPUTATION 
OF A INTERMEDIARY INDICATOR, CALLED HERE P_CRIT, WHICH IS A FUNCTION 
OF THE NUMBER OF VARIABLES "p" AND THE NUMBER OF ITEMS ANALYZED "n". 
THE FORMULAS, PROVIDED BY REINMANN, FILZMOSER AND GARRET, WERE OBTAINED 
BY THEM BY MEANS OF DATA SIMULATION
**********************************************************************/
/****************************************************************************
TO DISTINGUISH POTENTIAL OUTLIERS FROM EXTREME VALUES OF THE DATA GENERATING
PROCESS, THE THRESHOLD VALUE OBTAINED ABOVE IS COMPARED TO ANOTHER KEY VALUE,
CALLED HERE P_N_DELTA, COMPUTED BY COMPARING THE CHI-SQUARE DISTRIBUTION 
WITH THE EMPIRICAL DISTRIBUTION OF THE SQUARED MAHALANOBIS ROBUST DISTANCES
****************************************************************************/

Proc sort data = Robust_distance_detection_&cluster; 
by descending RD_DIST; run;

%global p_crit;

Data Gervini_p_crit;
if &p <= 10 then p_crit = (0.24 - 0.003*&p.)/SQRT(&n.);
else p_crit = (0.252 - 0.0018*&p.)/SQRT(&n.);
run;

Data _NULL_;
set Gervini_p_crit;
if _N_ = 1;
call symput('p_crit', p_crit);
run;

Proc sort data = Robust_distance_detection_&cluster; 
by RD_DIST; run; 

data hat_robust_m_v3;
set Robust_distance_detection_&cluster;
prb=(_n_ - 1)/&n.;
chiquant=cinv(prb, &p.);
if PROBCHI((RD_DIST)**2, &p) - prb > 0 and PROBCHI((RD_DIST)**2, &p) >= 0.975 
then P_N_DELTA = PROBCHI((RD_DIST)**2, &p) - prb;
else P_N_DELTA = 0;
/********************************************************************
Note: this computation takes into account only positive differences
because negative differences would not indicate presence of outliers
********************************************************************/
run;

data p_n_delta;
set hat_robust_m_v3;
if PROBCHI((RD_DIST)**2, &p) >= 0.975;
run;

%global Supremum_P_N_DELTA;

Proc SQL Noprint;
select MAX(P_N_DELTA) into :Supremum_P_N_DELTA
from p_n_delta;
quit;

%if &Supremum_P_N_DELTA > &p_crit %then %do;

Data Robust_adaptive_detection_&cluster;
set Hat_robust_m_v3;
if prb /* empirical distribution */ > 1 - &Supremum_P_N_DELTA 
then DISTRIBUTION_STATUS = "OUTLIER";
run;

/************************************************************************
data step re-assigns status of those providers previously classified as
EXTREME VALUE in the robust distance method, which does not distinguish
between EXTREME VALUES and OUTLIERS as the adaptive threshold method does
************************************************************************/
Data Robust_distance_detection_&cluster;
set Robust_distance_detection_&cluster;
if DISTRIBUTION_STATUS = "EXTREME VALUE" then DISTRIBUTION_STATUS = "OUTLIER";
run;

Proc sort data = Robust_adaptive_detection_&cluster; 
by descending RD_DIST; run;

%end;

%else %do;

Proc sort data = Robust_adaptive_detection_&cluster; 
by descending RD_DIST; run;

%end;

Proc sort data = Robust_distance_detection_&cluster; by descending RD_DIST; run;

%put &p_crit;
%mend Mdist;

/*------------------------------------------
here the analysis used a dataset partitioned
in 7 clusters and it included 2 variables
------------------------------------------*/
%Mdist(ALL, input_file, 321456, 2); 

%Mdist(1, input_file, 123456, 2); 

%Mdist(2, input_file, 431234, 2); 

%Mdist(3, input_file, 123456, 2); 

%Mdist(4, input_file, 780123, 2); 

%Mdist(5, input_file, 765432, 2); 

%Mdist(6, input_file, 567890, 2); 

%Mdist(7, input_file, 901234, 2); 
